import easynet
import time
import timer
import socket
import pickle

BUFFER_SIZE = 1024
DEFAULT_PORT = 6666

class Server(easynet.EasySocketHandler):
    def __init__(self, serverPort):
        self.serverPort = serverPort
        self.onlineUsers = {}        # map name->status
        self.userSockets = {}        # map name->EasySocket
        self.conversations = {}      # map convID->[name name name...]
        self.conversationCount = 0
        self.dispatchMap = {}
        self.dispatchMap["REGISTER"] = self.handleRegister
        self.dispatchMap["STATUS"] = self.handleStatus
        self.dispatchMap["INVITE"] = self.handleInvite
        self.dispatchMap["LIST_CONVERSATIONS"] = self.handleListConversations
        self.dispatchMap["JOIN"] = self.handleJoin
        self.dispatchMap["SEND_MESSAGE"] = self.handleSendMessage
        self.dispatchMap["LEAVE"] = self.handleLeave
        self.dispatchMap["GOODBYE"] = self.handleGoodbye

        self.server = easynet.EasyServer(serverPort, self.callback)

        self.updateTimer = timer.Timer(15, self.sendUpdates)

    def launch(self):
        print "+++ Server is running on port ", self.serverPort
        self.server.serve()
        self.updateTimer.start()

    # Invoked once every 15 seconds to send list of online users and their status strings
    def sendUpdates(self, timer):
        # send the entire dictionary of users
        # msg = "ONLINE_USERS " + pickle.dumps(self.onlineUsers)
        
        for user in self.userSockets.keys():
            self.sendUpdate(user)

    def callback(self, easySock, peerAddress, source):
        print "+++ Connected to client from ", peerAddress
        easySock.setHandler(self)

    def sendUpdate(self, user):
        easySock = self.userSockets[user]
	msg = "ONLINE_USERS " + pickle.dumps(self.onlineUsers)
        
	#print "sending"
        try:
            easySock.send(msg)
        except socket.error:
            # remove 'em if we can't contact 'em
            del(self.userSockets[user])
            del(self.onlineUsers[user])
	    print "contact error"

    def findUser(self, easySock):
        for user in self.userSockets.keys():
            if self.userSockets[user] == easySock:
                return user
        return None

    def handleIncomingMessage(self, data, easySock):
    	#print data
        idx = data.find(' ')
        if idx == -1:
            easySock.send("ERROR Badly formatted data: " + data)
            return
        command = data[0:idx]
        args = data[idx+1:]
        if not self.dispatchMap.has_key(command):
            easySock.send("ERROR No handler for " + command)
            return
        user = self.findUser(easySock)
        if user == None and command != "REGISTER":
            easySock.send("ERROR Not registered yet")
        handler = self.dispatchMap[command]
        handler(easySock, user, args)
        
    def handleSocketClosed(self, easySock):
        user = self.findUser(easySock)
        self.handleGoodbye(easySock, user, None)

    def handleRegister(self, easySock, user, args):
        try:
            data = pickle.loads(args)
            user = data[0]
            status = data[1]
            
	    print "+++ new user " + user
            self.userSockets[user] = easySock
            self.onlineUsers[user] = status
            
            #self.sendUpdate(user)
	    # Notify everyone
       	    for user in self.userSockets.keys():
            	self.sendUpdate(user)
        except IndexError:
            easySock.send("ERROR Badly formatted arguments (IndexError): " + args)           
        except KeyError:
            easySock.send("ERROR Badly formatted arguments (KeyError): " + args)
        except ValueError:
            easySock.send("ERROR Badly formatted arguments (ValueError): " + args)                       
    
        return 0

    def handleStatus(self, easySock, user, args):
        try:
            status = pickle.loads(args)
            
            self.onlineUsers[user] = status
        
        except IndexError:
            easySock.send("ERROR Badly formatted arguments (IndexError): " + args)
        except KeyError:
            easySock.send("ERROR Badly formatted arguments (KeyError): " + args)
        except ValueError:
            easySock.send("ERROR Badly formatted arguments (ValueError): " + args)                       
            
        return 0

    def handleInvite(self, easySock, user, args):
        try:
            invitees = pickle.loads(args)
	    #print invitees
        
            self.conversationCount = self.conversationCount + 1
            self.conversations[self.conversationCount] = [user]
            
            msg = "INVITATION " + pickle.dumps([self.conversationCount, user])
            for invitee in invitees:
                if self.userSockets.has_key(invitee):
                    inviteeSock = self.userSockets[invitee]
                    inviteeSock.send(msg)
                else:
                    easySock.send("ERROR Cannot find user " + invitee + " to invite to a conversation")

	    # Send an invitation back to the original sender, too
	    easySock.send(msg)
        except IndexError:
            easySock.send("ERROR Badly formatted arguments (IndexError): " + args)
        except KeyError:
            easySock.send("ERROR Badly formatted arguments (KeyError): " + args)
        except ValueError:
            easySock.send("ERROR Badly formatted arguments (ValueError): " + args)                       
        
        return 0

    def handleListConversations(self, easySock, user, args):
    	msg = pickle.dumps([self.conversations])
        easySock.send("CONVERSATIONS " + msg)
        
        return 0

    def handleJoin(self, easySock, user, args):
        try:
            data = pickle.loads(args)
	    convID = data[0]

	    #print self.conversations
          
            if self.conversations.has_key(convID):
                participants = self.conversations[convID]
		try:
		   participants.index(user)
		except ValueError:
                   participants.append(user)
                   self.conversations[convID] = participants

		   # send note to everyone involved
                   recipients = self.conversations[convID]
                   for recipient in recipients:
                     if self.userSockets.has_key(recipient):
                       recipientSock = self.userSockets[recipient]
                       recipientSock.send("CONVERSATIONS " + pickle.dumps([self.conversations]))
            else:
                easySock.send("ERROR Cannot find conversation " + args)            

        except IndexError:
            easySock.send("ERROR Badly formatted arguments (IndexError): " + args)
        except KeyError:
            easySock.send("ERROR Badly formatted arguments (KeyError): " + args)
        except ValueError:
            easySock.send("ERROR Badly formatted arguments (ValueError): " + args)
	except TypeError:
	    easySock.send("ERROR Badly formatted arguments (TypeError): " + args)
            
        return 0

    def handleSendMessage(self, easySock, user, args):
        try:
            data = pickle.loads(args)
            convID = data[0]
            msg = data[2]
            
            if self.conversations.has_key(convID):
                recipients = self.conversations[convID]
                for recipient in recipients:
                    if self.userSockets.has_key(recipient):
                        recipientSock = self.userSockets[recipient]
                        recipientSock.send("SEND_MESSAGE " + pickle.dumps([convID, user, msg]))
            else:
                easySock.send("ERROR Cannot find conversation " + args)

        except IndexError:
            easySock.send("ERROR Badly formatted arguments (IndexError): " + args)
        except KeyError:
            easySock.send("ERROR Badly formatted arguments (KeyError): " + args)
        except ValueError:
            easySock.send("ERROR Badly formatted arguments (ValueError): " + args)
	except TypeError:
	    easySock.send("ERROR Badly formatted arguments (TypeError): " + args)
        
        return 0

    def handleLeave(self, easySock, user, args):
        try:
            data = pickle.loads(args)
	    convID = data[0]           
 
            if self.conversations.has_key(convID):
                participants = self.conversations[convID]
                if user in participants:
                    participants.remove(user)
                    self.conversations[convID] = participants

		# Send everybody updated lists
                recipients = self.conversations[convID]
                for recipient in recipients:
                  if self.userSockets.has_key(recipient):
                    recipientSock = self.userSockets[recipient]
		    msg = pickle.dumps([self.conversations])
		    #print "msg is ",
		    #print len(msg)
                    recipientSock.send("CONVERSATIONS " + msg)
            else:
                easySock.send("ERROR Cannot find conversation " + args)

        except IndexError:
            easySock.send("ERROR Badly formatted arguments (IndexError): " + args)
        except KeyError:
            easySock.send("ERROR Badly formatted arguments (KeyError): " + args)
        except ValueError:
            easySock.send("ERROR Badly formatted arguments (ValueError): " + args)
	except TypeError:
	    easySock.send("ERROR Please send a list containing convID")
            
        return 0
        
    def handleGoodbye(self, easySock, user, args):
        if not user == None:
            if self.onlineUsers.has_key(user):
                del self.onlineUsers[user]
            if self.userSockets.has_key(user):
                del self.userSockets[user]
            for convID in self.conversations.keys():
                participants = self.conversations[convID]
                if user in participants:
                    participants.remove(user)
                    self.conversations[convID] = participants

		    # Send everybody updated lists
                    recipients = self.conversations[convID]
                    for recipient in recipients:
                      if self.userSockets.has_key(recipient):
                        recipientSock = self.userSockets[recipient]
                        recipientSock.send("CONVERSATIONS " + pickle.dumps([self.conversations]))
            
        return 1

if __name__ == "__main__":
    server = Server(DEFAULT_PORT)
    server.launch()

